@@ -5,6 +5,7 @@ __all__ = ['RemoteJIT', 'Signature', 'Caller']
 
 import os
 import inspect
+import warnings
 from . import irtools
 from .typesystem import Type, get_signature
 from .thrift import Server, Dispatcher, dispatchermethod, Data, Client
@@ -20,6 +21,44 @@ def isfunctionlike(obj):
     return True
 
 
+def extract_templates(options):
+    """Extract templates mapping data from options dictionary.
+
+    If options does not contain "templates", it will be constructed
+    from all unknown options that have list values. Otherwise, the
+    corresponding value is returned with no further processing of
+    options content.
+
+    Parameters
+    ----------
+    options : dict
+
+    Returns
+    -------
+    options : dict
+      A copy of input without templates mapping data.
+    templates : dict
+      Templates mapping which is a collections of pairs of template
+      name and a list of concrete types. Template name cannot
+      correspond to a concrete type.
+
+    """
+    known_options = ['devices', 'local']
+    new_options = {}
+    templates = options.get('templates')
+    if templates is not None:
+        new_options.update(options)
+        del new_options['templates']
+    else:
+        templates = {}
+        for k, v in options.items():
+            if (isinstance(k, str) and isinstance(v, list) and k not in known_options):
+                templates[k] = v
+            else:
+                new_options[k] = v
+    return new_options, templates
+
+
 class Signature(object):
     """Signature decorator for Python functions.
 
@@ -55,18 +94,26 @@ class Signature(object):
         assert isinstance(remotejit, RemoteJIT), type(remotejit)
         self.remotejit = remotejit
         self.signatures = []
+        self.signature_devices = {}
+        self.signature_templates = {}
+
+    @property
+    def debug(self):
+        return self.remotejit.debug
 
     @property
     def local(self):
         sig = Signature(self.remotejit.local)
         sig.signatures.extend(self.signatures)
+        assert not self.signature_devices
+        assert not self.signature_templates
         return sig
 
     def __str__(self):
         lst = ["'%s'" % (s,) for s in self.signatures]
         return '%s(%s)' % (self.__class__.__name__, ', '.join(lst))
 
-    def __call__(self, obj):
+    def __call__(self, obj, **options):
         """Decorate signatures or a function.
 
         Parameters
@@ -74,6 +121,13 @@ class Signature(object):
         obj : {str, Signature, function, ...}
           Specify object that represents a function type.
 
+        Keyword parameters
+        ------------------
+        devices : list
+          Specify device names for the given set of signatures.
+        templates : dict
+          Specify template types mapping.
+
         Returns
         -------
         result : {Signature, Caller}
@@ -87,13 +141,24 @@ class Signature(object):
         etc) depend on the target device which information will be
         available at the compile stage. The target dependent
         signatures can be retrieved using
-        `signature.get_signatures(target_info)`.
+        `signature.get_signatures()`.
         """
         if obj is None:
             return self
+        options, templates = extract_templates(options)
+        devices = options.get('devices')
         if isinstance(obj, Signature):
             self.signatures.extend(obj.signatures)
+            self.signature_devices.update(obj.signature_devices)
             self.remotejit.discard_last_compile()
+            if devices is not None:
+                for s in obj.signatures:
+                    self.signature_devices[s] = devices
+            assert not templates
+            for s in obj.signatures:
+                t = obj.signature_templates.get(s)
+                if t is not None:
+                    self.signature_templates[s] = t
             return self
         if isinstance(obj, Caller):
             # return new Caller with extended signatures set
@@ -101,16 +166,24 @@ class Signature(object):
             final = Signature(self.remotejit)
             final(self)  # copies the signatures from self to final
             final(obj.signature)  # copies the signatures from obj to final
+            assert devices is None
+            assert not templates
             return Caller(obj.func, final)
         if isfunctionlike(obj):
             final = Signature(self.remotejit)
             final(self)  # copies the signatures from self to final
+            assert devices is None
+            assert not templates
             return Caller(obj, final)
         self.signatures.append(obj)
         self.remotejit.discard_last_compile()
+        if devices is not None:
+            self.signature_devices[obj] = devices
+        if templates:
+            self.signature_templates[obj] = templates
         return self
 
-    def best_match(self, func, atypes: tuple, target_info: TargetInfo) -> Type:
+    def best_match(self, func, atypes: tuple) -> Type:
         """Return function type from signatures that matches best with given
         argument types.
 
@@ -120,8 +193,6 @@ class Signature(object):
         ----------
         atypes : Type-tuple
           Specify a tuple of argument types.
-        target_info : TargetInfo
-          Specify target device information.
 
         Returns
         -------
@@ -131,7 +202,7 @@ class Signature(object):
         """
         ftype = None
         match_penalty = None
-        available_types = self.normalized(func, target_info).signatures
+        available_types = self.normalized(func).signatures
         for typ in available_types:
             penalty = typ.match(atypes)
             if penalty is not None:
@@ -146,42 +217,57 @@ class Signature(object):
                 f' `{satypes}`. Available function types: {available}')
         return ftype
 
-    def normalized(self, func, target_info: TargetInfo):
+    def normalized(self, func=None):
         """Return a copy of Signature object where all signatures are
-        normalized to Type instances using given target device
+        normalized to Type instances using the current target device
         information.
 
         Parameters
         ----------
-        func : function
+        func : {None, callable}
           Python function that annotations are attached to signature.
-        target_info : TargetInfo
 
         Returns
         -------
         signature : Signature
         """
         signature = Signature(self.remotejit)
-        fsig = Type.fromcallable(func, target_info)
-        nargs = len(fsig[1])
+        fsig = Type.fromcallable(func) if func is not None else None
+        nargs = fsig.arity if func is not None else None
+        target_info = TargetInfo()
         for sig in self.signatures:
-            if nargs is None:
-                nargs = len(sig[1])
-            sig = Type.fromobject(sig, target_info)
+            devices = self.signature_devices.get(sig)
+            if not target_info.check_enabled(devices):
+                if self.debug:
+                    print(f'{type(self).__name__}.normalized: skipping {sig} as'
+                          f' not supported by devices: {devices}')
+                continue
+            templates = self.signature_templates.get(sig, {})
+            sig = Type.fromobject(sig)
             if not sig.is_complete:
+                warnings.warn(f'Incomplete signature {sig} will be ignored')
                 continue
             if not sig.is_function:
                 raise ValueError(
                     'expected signature representing function type,'
                     f' got `{sig}`')
-            if len(sig[1]) != nargs:
+            if nargs is None:
+                nargs = sig.arity
+            elif sig.arity != nargs:
                 raise ValueError(f'signature `{sig}` must have arity {nargs}'
                                  f' but got {len(sig[1])}')
             if fsig is not None:
                 sig.inherit_annotations(fsig)
-            if sig not in signature.signatures:
-                signature.signatures.append(sig)
-        if fsig.is_complete:
+
+            if not sig.is_concrete:
+                for csig in sig.apply_templates(templates):
+                    assert isinstance(csig, Type), (sig, csig, type(csig))
+                    if csig not in signature.signatures:
+                        signature.signatures.append(csig)
+            else:
+                if sig not in signature.signatures:
+                    signature.signatures.append(sig)
+        if fsig is not None and fsig.is_complete:
             if fsig not in signature.signatures:
                 signature.signatures.append(fsig)
         return signature
@@ -208,6 +294,7 @@ class Caller(object):
         """
         self.remotejit = signature.remotejit
         self.signature = signature
+        func = self.remotejit.preprocess_callable(func)
         self.func = func
         self.nargs = len(get_signature(func).parameters)
 
@@ -235,21 +322,27 @@ class Caller(object):
         """Return LLVM IRs of all target devices.
         """
         lst = ['']
+        fid = 0
         for device, target_info in self.remotejit.targets.items():
-            lst.append(f'{device:-^80}')
-            signatures = self.get_signatures(target_info)
-            llvm_module = irtools.compile_to_LLVM(
-                [(self.func, signatures)],
-                target_info,
-                debug=self.remotejit.debug)
-            lst.append(str(llvm_module))
+            with target_info:
+                lst.append(f'{device:-^80}')
+                signatures = self.get_signatures()
+                signatures_map = {}
+                for sig in signatures:
+                    fid += 1
+                    signatures_map[fid] = sig
+                llvm_module, succesful_fids = irtools.compile_to_LLVM(
+                    [(self.func, signatures_map)],
+                    target_info,
+                    debug=self.remotejit.debug)
+                lst.append(str(llvm_module))
         lst.append(f'{"":-^80}')
         return '\n'.join(lst)
 
-    def get_signatures(self, target_info: TargetInfo):
+    def get_signatures(self):
         """Return a list of normalized signatures for given target device.
         """
-        return self.signature.normalized(self.func, target_info).signatures
+        return self.signature.normalized(self.func).signatures
 
     # RBC user-interface
 
@@ -265,13 +358,14 @@ class Caller(object):
                     f' one device. Available devices: {", ".join(targets)}')
             device = tuple(targets)[0]
         target_info = targets[device]
-        atypes = tuple(Type.fromvalue(a, target_info) for a in arguments)
-        ftype = self.signature.best_match(self.func, atypes, target_info)
-        key = self.func.__name__, ftype
-        if key not in self._is_compiled:
-            self.remotejit.remote_compile(self.func, ftype, target_info)
-            self._is_compiled.add(key)
-        return self.remotejit.remote_call(self.func, ftype, arguments)
+        with target_info:
+            atypes = tuple(map(Type.fromvalue, arguments))
+            ftype = self.signature.best_match(self.func, atypes)
+            key = self.func.__name__, ftype
+            if key not in self._is_compiled:
+                self.remotejit.remote_compile(self.func, ftype, target_info)
+                self._is_compiled.add(key)
+            return self.remotejit.remote_call(self.func, ftype, arguments)
 
 
 class RemoteJIT(object):
@@ -301,8 +395,6 @@ class RemoteJIT(object):
     The sum will be evaluated in the remote host.
     """
 
-    converters = []
-
     multiplexed = True
 
     thrift_content = None
@@ -438,10 +530,7 @@ class RemoteJIT(object):
         """Return device-target_info mapping of the remote server.
         """
         if self._targets is None:
-            self._targets = targets = self.retrieve_targets()
-            for target in targets.values():
-                for c in self.converters:
-                    target.add_converter(c)
+            self._targets = self.retrieve_targets()
         return self._targets
 
     def __call__(self, *signatures, **options):
@@ -453,7 +542,14 @@ class RemoteJIT(object):
           Specify signatures of a remote JIT function, or a Python
           function as a template from which the remote JIT function
           will be compiled.
+
+        Keyword parameters
+        ------------------
         local : bool
+        devices : list
+          Specify device names for the given set of signatures.
+        templates : dict
+          Specify template types mapping.
 
         Returns
         -------
@@ -473,8 +569,10 @@ class RemoteJIT(object):
             s = Signature(self.local)
         else:
             s = Signature(self)
+        devices = options.get('devices')
+        options, templates = extract_templates(options)
         for sig in signatures:
-            s = s(sig)
+            s = s(sig, devices=devices, templates=templates)
         return s
 
     def start_server(self, background=False):
@@ -517,7 +615,7 @@ class RemoteJIT(object):
                 port=self.port,
                 multiplexed=self.multiplexed,
                 thrift_content=self.thrift_content,
-                socket_timeout=10000)
+                socket_timeout=60000)
         return self._client
 
     def remote_compile(self, func, ftype: Type, target_info: TargetInfo):
@@ -530,8 +628,8 @@ class RemoteJIT(object):
         Return the corresponding LLVM IR module instance which may be
         useful for debugging.
         """
-        llvm_module = irtools.compile_to_LLVM(
-            [(func, [ftype])], target_info, debug=self.debug)
+        llvm_module, succesful_fids = irtools.compile_to_LLVM(
+            [(func, {0: ftype})], target_info, debug=self.debug)
         ir = str(llvm_module)
         mangled_signatures = ';'.join([s.mangle() for s in [ftype]])
         response = self.client(remotejit=dict(
@@ -551,6 +649,20 @@ class RemoteJIT(object):
         response = self.client(remotejit=dict(call=(fullname, arguments)))
         return response['remotejit']['call']
 
+    def preprocess_callable(self, func):
+        """Preprocess func to be used as a remotejit function definition.
+
+        Parameters
+        ----------
+        func : callable
+
+        Returns
+        -------
+        func : callable
+          Preprocessed func.
+        """
+        return func
+
 
 class DispatcherRJIT(Dispatcher):
     """Implements remotejit service methods.
