@@ -52,6 +52,12 @@ use_host_target: False
             os.environ['OMNISCI_CLIENT_CONF'] = old_conf
 
 
+@pytest.fixture(scope='module')
+def nb_version():
+    from rbc.utils import get_version
+    return get_version('numba')
+
+
 @pytest.fixture(scope='module')
 def omnisci():
     config = rbc_omnisci.get_client_config(debug=not True)
@@ -175,6 +181,44 @@ def test_single_argument_overloading(omnisci):
         assert isinstance(x1, type(x))
 
 
+def test_numpy_forbidden_ufunc(omnisci, nb_version):
+    omnisci.reset()
+
+    if omnisci.version >= (5, 5) and nb_version >= (0, 52):
+        pytest.skip(
+            f'forbidden ufunc not required for OmniSciDB {omnisci.version} '
+            f'and Numba {nb_version}')
+
+    if not omnisci.has_cuda:
+        pytest.skip('forbidden ufunc not required for CUDA-disabled OmniSciDB')
+
+    if omnisci.version >= (5, 5):
+        # Requires: https://github.com/omnisci/omniscidb-internal/pull/4955
+        pytest.skip('forbidden ufunc not required if CPU ufuncs work [omniscidb-interal PR 4955]')
+
+    msg = "Attempt to use function with name `{ufunc}`"
+
+    @omnisci('double(double)')
+    def arcsin(x):
+        return np.arcsin(x)
+
+    with pytest.raises(errors.ForbiddenIntrinsicError) as excinfo:
+        omnisci.register()
+    assert msg.format(ufunc='asin') in str(excinfo.value)
+
+    omnisci.reset()
+
+    @omnisci('float32(float32, float32)')
+    def logaddexp(x, y):
+        return np.logaddexp(x, y)
+
+    with pytest.raises(errors.ForbiddenIntrinsicError) as excinfo:
+        omnisci.register()
+    assert msg.format(ufunc='log1pf') in str(excinfo.value)
+
+    omnisci.reset()
+
+
 def test_thrift_api_doc(omnisci):
     omnisci.reset()
 
@@ -203,7 +247,6 @@ def test_manual_ir(omnisci):
                       (4.0, 4.0, 4, 4, 4, 4, 1)]
     device_params = omnisci.thrift_call('get_device_parameters',
                                         omnisci.session_id)
-    # print(device_params)
     cpu_target_triple = device_params['cpu_triple']
     cpu_target_datalayout = "e-m:e-i64:64-f80:128-n8:16:32:64-S128"
     gpu_target_triple = device_params.get('gpu_triple')
@@ -271,8 +314,9 @@ entry:
                             ast_signatures, device_ir_map)
 
 
-@pytest.mark.skip(reason='omnisci server crashes')
 def test_ir_query_error(omnisci):
+    pytest.skip("requires omniscidb-internal catching undefined symbols")
+
     device_params = omnisci.thrift_call('get_device_parameters',
                                         omnisci.session_id)
     gpu_target_triple = device_params.get('gpu_triple')
@@ -292,9 +336,12 @@ entry:
 
     omnisci.thrift_call('register_runtime_udf', omnisci.session_id,
                         ast_signatures, device_ir_map)
-    descr, result = omnisci.sql_execute(
-        'SELECT i4, foobar(i4, i4) FROM {omnisci.table_name}'
-        .format(**locals()))
+    try:
+        omnisci.sql_execute(
+            'SELECT i4, foobar(i4, i4) FROM {omnisci.table_name}'
+            .format(**locals()))
+    except errors.OmnisciServerError as msg:
+        assert "use of undefined value '@foobar'" in str(msg)
 
 
 def test_multiple_implementation(omnisci):
@@ -566,7 +613,8 @@ def test_casting(omnisci):
     ]:
         q = ('select '+f+'('+v+') from {omnisci.table_name} limit 1'
              .format(**locals()))
-        match = r".*Function "+f+r"\("+t+r"\) not supported"
+        match = (r".*(Function "+f+r"\("+t+r"\) not supported"
+                 r"|Could not bind "+f+r"\("+t+r"\))")
         with pytest.raises(Exception, match=match):
             descr, result = omnisci.sql_execute(q)
             print('query: ', q)
@@ -591,7 +639,8 @@ def test_casting(omnisci):
                 continue
             with pytest.raises(
                     Exception,
-                    match=r".*Function "+f+r"\("+at+r"\) not supported"):
+                    match=(r".*(Function "+f+r"\("+at+r"\) not supported"
+                           r"|Could not bind "+f+r"\("+at+r"\))")):
                 descr, result = omnisci.sql_execute(
                     'select '+f+'('+av+') from {omnisci.table_name} limit 1'
                     .format(**locals()))
@@ -605,7 +654,8 @@ def test_casting(omnisci):
         ]:
             with pytest.raises(
                     Exception,
-                    match=r".*Function "+f+r"\("+at+r"\) not supported"):
+                    match=(r".*(Function "+f+r"\("+at+r"\) not supported"
+                           r"|Could not bind "+f+r"\("+at+r"\))")):
                 descr, result = omnisci.sql_execute(
                     'select '+f+'('+av+') from {omnisci.table_name} limit 1'
                     .format(**locals()))
