@@ -1,13 +1,13 @@
 try:
     import numba as nb
-    nb_NA_message = None
+    nb_NA_message = ''
 except ImportError as msg:
     nb = None
     nb_NA_message = str(msg)
 
 try:
     import numpy as np
-    np_NA_message = None
+    np_NA_message = ''
 except ImportError as msg:
     np = None
     np_NA_message = str(msg)
@@ -22,23 +22,13 @@ from rbc.typesystem import Type, get_signature
 from rbc.utils import get_datamodel
 from rbc.targetinfo import TargetInfo
 
-target_info = TargetInfo.host()
 
+@pytest.fixture(scope='module')
+def target_info():
 
-def Type_fromstring(s):
-    return Type.fromstring(s, target_info)
-
-
-def Type_fromobject(s):
-    return Type.fromobject(s, target_info)
-
-
-def Type_fromcallable(s):
-    return Type.fromcallable(s, target_info)
-
-
-def Type_fromvalue(s):
-    return Type.fromvalue(s, target_info)
+    target_info = TargetInfo.host()
+    with target_info:
+        yield target_info
 
 
 def test_findparen():
@@ -66,82 +56,88 @@ def test_commasplit():
     assert '^'.join(commasplit('a[:, :], b[:, :, :]')) == 'a[:, :]^b[:, :, :]'
 
 
-def test_fromstring():
-    assert Type_fromstring('void') == Type()
-    assert Type_fromstring('') == Type()
-    assert Type_fromstring('none') == Type()
-    assert Type_fromstring('i') == Type('int32')
-    assert Type_fromstring('i*') == Type(Type('int32'), '*')
-    assert Type_fromstring('*') == Type(Type(), '*')
-    assert Type_fromstring('void*') == Type(Type(), '*')
-    assert Type_fromstring('{i,j}') == Type(Type('int32'), Type('j'))
-    assert Type_fromstring('i(j)') == Type(Type('int32'), (Type('j'), ))
-    assert Type_fromstring('i(j , k)') == Type(Type('int32'),
+def test_fromstring(target_info):
+
+    assert Type.fromstring('void') == Type()
+    assert Type.fromstring('') == Type()
+    assert Type.fromstring('none') == Type()
+    assert Type.fromstring('i') == Type('int32')
+    assert Type.fromstring('i*') == Type(Type('int32'), '*')
+    assert Type.fromstring('*') == Type(Type(), '*')
+    assert Type.fromstring('void*') == Type(Type(), '*')
+    assert Type.fromstring('{i,j}') == Type(Type('int32'), Type('j'))
+    assert Type.fromstring('i(j)') == Type(Type('int32'), (Type('j'), ))
+    assert Type.fromstring('i(j , k)') == Type(Type('int32'),
                                                (Type('j'), Type('k')))
-    assert Type_fromstring('  (j , k) ') == Type(Type(),
+    assert Type.fromstring('  (j , k) ') == Type(Type(),
                                                  (Type('j'), Type('k')))
-    assert Type_fromstring('void(j,k)') == Type(Type(),
+    assert Type.fromstring('void(j,k)') == Type(Type(),
                                                 (Type('j'), Type('k')))
 
-    assert Type_fromstring('i a') == Type('int32', name='a')
-    assert Type_fromstring('i* a') == Type(Type('int32'), '*', name='a')
-    assert Type_fromstring('{i,j} a') == Type(Type('int32'), Type('j'),
+    assert Type.fromstring('i a') == Type('int32', name='a')
+    assert Type.fromstring('i* a') == Type(Type('int32'), '*', name='a')
+    assert Type.fromstring('{i,j} a') == Type(Type('int32'), Type('j'),
                                               name='a')
-    assert Type_fromstring('i(j) a') == Type(Type('int32'), (Type('j'),),
+    assert Type.fromstring('i(j) a') == Type(Type('int32'), (Type('j'),),
                                              name='a')
-    assert Type_fromstring('i a*') == Type(Type('int32', name='a'), '*')
-    assert Type_fromstring('{i a,j b} c') == Type(Type('int32', name='a'),
+    assert Type.fromstring('i a*') == Type(Type('int32', name='a'), '*')
+    assert Type.fromstring('{i a,j b} c') == Type(Type('int32', name='a'),
                                                   Type('j', name='b'),
                                                   name='c')
+    assert Type.fromstring('A<i>') == Type(('A', Type('int32')))
+    assert Type.fromstring('A<i*>') == Type(('A', Type(Type('int32'), '*')))
+    assert Type.fromstring('A<i>*') == Type(Type(('A', Type('int32'),)), '*')
 
     with pytest.raises(ValueError, match=r'failed to find lparen index in'):
-        Type_fromstring('a)')
+        Type.fromstring('a)')
 
     with pytest.raises(ValueError, match=r'failed to comma-split'):
-        Type_fromstring('a((b)')
+        Type.fromstring('a((b)')
 
     with pytest.raises(ValueError, match=r'failed to comma-split'):
-        Type_fromstring('a((b)')
+        Type.fromstring('a((b)')
 
     with pytest.raises(ValueError, match=r'mismatching curly parenthesis in'):
-        Type_fromstring('ab}')
+        Type.fromstring('ab}')
 
 
-def test_is_properties():
+def test_is_properties(target_info):
     t = Type()
     assert t._is_ok and t.is_void
     t = Type('i')
     assert t._is_ok and t.is_atomic
     t = Type('ij')
     assert t._is_ok and t.is_atomic
-    t = Type_fromstring('ij')
+    t = Type.fromstring('ij')
     assert t._is_ok and t.is_atomic
-    t = Type_fromstring('i,j')
+    t = Type.fromstring('i,j')
     assert t._is_ok and t.is_atomic  # !
-    t = Type_fromstring('i  *')
+    t = Type.fromstring('i  *')
     assert t._is_ok and t.is_pointer
-    t = Type_fromstring('*')
+    t = Type.fromstring('*')
     assert t._is_ok and t.is_pointer
-    t = Type_fromstring('i* * ')
+    t = Type.fromstring('i* * ')
     assert t._is_ok and t.is_pointer
-    t = Type_fromstring('i(j)')
+    t = Type.fromstring('i(j)')
     assert t._is_ok and t.is_function
-    t = Type_fromstring('(j)')
+    t = Type.fromstring('(j)')
     assert t._is_ok and t.is_function
-    t = Type_fromstring('()')
+    t = Type.fromstring('()')
     assert t._is_ok and t.is_function
-    t = Type_fromstring('{i, j}')
+    t = Type.fromstring('{i, j}')
     assert t._is_ok and t.is_struct
+    t = Type.fromstring('A<i>')
+    assert t._is_ok and t.is_custom
 
     with pytest.raises(ValueError,
                        match=r'attempt to create an invalid Type object from'):
         Type('a', 'b')
 
 
-def test_tostring():
+def test_tostring(target_info):
 
     def tostr(a):
-        return Type_fromstring(a).tostring()
+        return Type.fromstring(a).tostring()
 
     assert tostr('a') == 'a'
     assert tostr('()') == 'void(void)'
@@ -157,12 +153,13 @@ def test_tostring():
     assert tostr('*') == 'void*'
     assert tostr('void *') == 'void*'
     assert tostr('*(*,{*,*})') == 'void*(void*, {void*, void*})'
+    assert tostr('A<a>') == 'A<a>'
 
 
-def test_normalize():
+def test_normalize(target_info):
 
     def tostr(a):
-        return Type_fromstring(a).tostring()
+        return Type.fromstring(a).tostring()
 
     assert tostr('a') == 'a'
     assert tostr('int32') == 'int32'
@@ -203,6 +200,7 @@ def test_normalize():
     assert tostr('i(i*, i15)') == 'int32(int32*, int15)'
     assert tostr('{i,d,c, bool,f,str*}') \
         == '{int32, float64, complex64, bool, float32, string*}'
+    assert tostr('A<i(B<f64>)>') == 'A<int32(B<float64>)>'
 
     datamodel = get_datamodel()
     if datamodel == 'LP64':
@@ -249,11 +247,11 @@ def test_normalize():
         raise NotImplementedError('tests for datamodel=`%s`' % (datamodel))
 
 
-def test_toctypes():
+def test_toctypes(target_info):
     import ctypes
 
     def toctypes(a):
-        return Type_fromstring(a).toctypes()
+        return Type.fromstring(a).toctypes()
 
     assert toctypes('bool') == ctypes.c_bool
     assert toctypes('i8') == ctypes.c_int8
@@ -276,14 +274,14 @@ def test_toctypes():
     assert s._fields_ == [('f0', ctypes.c_int32), ('f1', ctypes.c_double)]
 
 
-def test_fromctypes():
+def test_fromctypes(target_info):
     import ctypes
 
     def fromstr(a):
-        return Type_fromstring(a)
+        return Type.fromstring(a)
 
     def fromctypes(t):
-        return Type.fromctypes(t, target_info)
+        return Type.fromctypes(t)
 
     assert fromctypes(ctypes.c_char_p) == fromstr('char*')
     assert fromctypes(ctypes.c_wchar_p) == fromstr('wchar*')
@@ -305,9 +303,9 @@ def test_fromctypes():
 
 
 @pytest.mark.skipif(nb is None, reason=nb_NA_message)
-def test_tonumba():
+def test_tonumba(target_info):
     def tonumba(a):
-        return Type_fromstring(a).tonumba()
+        return Type.fromstring(a).tonumba()
 
     assert tonumba('void') == nb.void
     assert tonumba('bool') == nb.boolean
@@ -330,14 +328,14 @@ def test_tonumba():
 
 
 @pytest.mark.skipif(nb is None, reason=nb_NA_message)
-def test_fromnumba():
+def test_fromnumba(target_info):
     import numba as nb
 
     def fromstr(a):
-        return Type_fromstring(a)
+        return Type.fromstring(a)
 
     def fromnumba(t):
-        return Type.fromnumba(t, target_info)
+        return Type.fromnumba(t)
 
     assert fromnumba(nb.void) == fromstr('void')
     assert fromnumba(nb.boolean) == fromstr('bool')
@@ -358,13 +356,13 @@ def test_fromnumba():
 
 
 @pytest.mark.skipif(np is None, reason='NumPy is not available')
-def test_fromnumpy():
+def test_fromnumpy(target_info):
 
     def fromstr(a):
-        return Type_fromstring(a)
+        return Type.fromstring(a)
 
     def fromnumpy(t):
-        return Type.fromnumpy(t, target_info)
+        return Type.fromnumpy(t)
 
     assert fromnumpy(np.void) == fromstr('void')
     assert fromnumpy(np.bool_) == fromstr('bool')
@@ -396,63 +394,63 @@ def test_fromnumpy():
         assert fromnumpy(np.complex256) == fromstr('complex256')
 
 
-def test_fromcallable():
+def test_fromcallable(target_info):
 
     def foo(a: int, b: float) -> int:
         pass
 
-    assert Type_fromcallable(foo) == Type_fromstring('i64(i64,d)')
+    assert Type.fromcallable(foo) == Type.fromstring('i64(i64,d)')
 
     def foo(a: 'int32', b):  # noqa: F821
         pass
 
-    assert Type_fromcallable(foo) == Type_fromstring('void(i32,<type of b>)')
+    assert Type.fromcallable(foo) == Type.fromstring('void(i32,<type of b>)')
 
     with pytest.raises(
             ValueError,
             match=(r'constructing Type instance from'
                    r' a lambda function is not supported')):
-        Type_fromcallable(lambda a: a)
+        Type.fromcallable(lambda a: a)
 
     with pytest.raises(
             ValueError,
             match=r'callable argument kind must be positional'):
         def foo(*args): pass
-        Type_fromcallable(foo)
+        Type.fromcallable(foo)
 
 
-def test_fromvalue():
-    assert Type_fromvalue(1) == Type_fromstring('i64')
-    assert Type_fromvalue(1.0) == Type_fromstring('f64')
-    assert Type_fromvalue(1j) == Type_fromstring('c128')
-    assert Type_fromvalue("123".encode()) == Type_fromstring('char*')
-    assert Type_fromvalue("123") == Type_fromstring('string')
+def test_fromvalue(target_info):
+    assert Type.fromvalue(1) == Type.fromstring('i64')
+    assert Type.fromvalue(1.0) == Type.fromstring('f64')
+    assert Type.fromvalue(1j) == Type.fromstring('c128')
+    assert Type.fromvalue("123".encode()) == Type.fromstring('char*')
+    assert Type.fromvalue("123") == Type.fromstring('string')
     x = np.dtype(np.float64).type(3.0)
-    assert Type_fromvalue(x) == Type_fromstring('float64')
+    assert Type.fromvalue(x) == Type.fromstring('float64')
     y = np.dtype(np.complex64).type((1+2j))
-    assert Type_fromvalue(y) == Type_fromstring('complex64')
+    assert Type.fromvalue(y) == Type.fromstring('complex64')
 
 
-def test_fromobject():
+def test_fromobject(target_info):
     import ctypes
-    assert Type_fromobject('i8') == Type_fromstring('i8')
-    assert Type_fromobject(int) == Type_fromstring('i64')
-    assert Type_fromobject(ctypes.c_int16) == Type_fromstring('i16')
+    assert Type.fromobject('i8') == Type.fromstring('i8')
+    assert Type.fromobject(int) == Type.fromstring('i64')
+    assert Type.fromobject(ctypes.c_int16) == Type.fromstring('i16')
     if nb is not None:
-        assert Type_fromobject(nb.int16) == Type_fromstring('i16')
+        assert Type.fromobject(nb.int16) == Type.fromstring('i16')
     if np is not None:
-        assert Type_fromobject(np.int32) == Type_fromstring('i32')
-        assert Type_fromobject(np.complex64) == Type_fromstring('complex64')
+        assert Type.fromobject(np.int32) == Type.fromstring('i32')
+        assert Type.fromobject(np.complex64) == Type.fromstring('complex64')
 
     def foo():
         pass
 
-    assert Type_fromobject(foo) == Type_fromstring('void(void)')
+    assert Type.fromobject(foo) == Type.fromstring('void(void)')
 
 
-def test_mangling():
+def test_mangling(target_info):
     def check(s):
-        t1 = Type_fromstring(s)
+        t1 = Type.fromstring(s)
         m = t1.mangle()
         try:
             t2 = Type.demangle(m)
@@ -469,9 +467,10 @@ def test_mangling():
     random_types = ['i8', 'bool', 'f', 'd', '{f}', '{f,d}', '{{f},d}',
                     '{f,{d}}', '{{f},{d}}', '{{{{f}}}}', '()', 'f(d)',
                     'f(())', 'f(d(f))', 'f(d)()', 'f(d)(f(d,d))',
-                    '{f}()', '{f,d}({f},f(d,d,d))']
+                    '{f}()', '{f,d}({f},f(d,d,d))', 'A<d, d>']
     unknown_types = ['a', 'a*', 'a()', '(a)', 'a(a)', '{a}', '({a})', '{a,a}',
-                     'foo', 'bar123', 'V', 'VVV', '_abc_', '_', 'A', 'K', 'P']
+                     'foo', 'bar123', 'V', 'VVV', '_abc_', '_', 'A', 'K', 'P',
+                     'A<a>']
     for s in atomic_types + random_types + unknown_types:
         check(s)
         check(s+'*')
@@ -488,17 +487,17 @@ def test_mangling():
     check('()')
 
 
-def test_unspecified():
-    assert str(Type_fromstring('unknown(_0,_1)')) == 'unknown(_0, _1)'
+def test_unspecified(target_info):
+    assert str(Type.fromstring('unknown(_0,_1)')) == 'unknown(_0, _1)'
 
 
-def test_annotation():
-    t = Type_fromstring('int foo| a = 1')
+def test_annotation(target_info):
+    t = Type.fromstring('int foo| a = 1')
     assert t.annotation() == dict(a='1')
     assert t[0] == 'int32'
 
     def tostr(a):
-        return Type_fromstring(a).tostring()
+        return Type.fromstring(a).tostring()
 
     assert tostr('int foo| a = 1') == 'int32 foo | a=1'
     assert tostr('int foo| a = 1 | b') == 'int32 foo | a=1 | b'
@@ -513,13 +512,13 @@ def test_annotation():
     assert tostr('int foo|a') == 'int32 foo | a'
     assert tostr('int foo|=1') == 'int32 foo | =1'
 
-    t = Type_fromstring('int')
+    t = Type.fromstring('int')
     assert (t | 'a').tostring() == 'int32 | a'
     assert (t | dict(b=1, c=2)).tostring() == 'int32 | b=1 | c=2'
 
 
 @pytest.mark.skipif(np is None, reason=np_NA_message)
-def test_get_signature_ufunc():
+def test_get_signature_ufunc(target_info):
 
     # Make sure that all get_signature can be applied to all numpy
     # ufuncs
