@@ -2,6 +2,7 @@ import socket
 import subprocess
 import ipaddress
 import netifaces
+import re
 import uuid
 import ctypes
 import llvmlite.binding as llvm
@@ -17,6 +18,41 @@ def get_version(package):
     raise NotImplementedError(f'get version of package {package}')
 
 
+def parse_version(version):
+    """Return parsed version tuple from version string.
+
+    For instance:
+
+    >>> parse_version('1.2.3dev4')
+    (1, 2, 3, 'dev4')
+    """
+
+    m = re.match(r'(\d+)[.](\d+)[.](\d+)(.*)', version)
+    if m is not None:
+        major, minor, micro, dev = m.groups()
+        if not dev:
+            return (int(major), int(minor), int(micro))
+        return (int(major), int(minor), int(micro), dev)
+
+    m = re.match(r'(\d+)[.](\d+)(.*)', version)
+    if m is not None:
+        major, minor, dev = m.groups()
+        if not dev:
+            return (int(major), int(minor))
+        return (int(major), int(minor), dev)
+
+    m = re.match(r'(\d+)(.*)', version)
+    if m is not None:
+        major, dev = m.groups()
+        if not dev:
+            return int(major),
+        return (int(major), dev)
+
+    if version:
+        return version,
+    return ()
+
+
 def get_local_ip():
     """Return localhost IP.
     """
